/*
 * loopup_udisks.h - Header file for lookup_udisks automount module
 *
 *   Copyright 2012 SuSE LINUX Products GmbH - All Rights Reserved
 *   Copyright 2012 Werner Fink <werner@suse.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
 *   USA; either version 2 of the License, or (at your option) any later
 *   version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 */

#ifndef LOOKUP_UDISKS_H
#define LOOKUP_UDISKS_H

#include "list.h"
typedef struct list_head list_t;

#ifndef list_for_each_safe
#define list_for_each_safe(pos, safe, head)	\
	for (pos = (head)->next, safe = pos->next; pos != (head); pos = safe, safe = pos->next)
#endif
#ifdef offsetof
# undef list_entry
# define list_entry(ptr, type, member)		(__extension__ ({	\
	__const__ __typeof__( ((type*)0)->member ) *__mptr = (ptr);	\
	((type *)( (char *)__mptr - offsetof(type,member) ));}))
#endif

#if !defined(__STDC_VERSION__) || (__STDC_VERSION__ < 199901L)
# ifndef  inline
#  define inline		__inline__
# endif
# ifndef  restrict
#  define restrict		__restrict__
# endif
# ifndef  volatile
#  define volatile		__volatile__
# endif
# ifndef  extension
#  define extension		__extension__
# endif
# ifndef  typeof
#  define typeof		__typeof__
# endif
#endif
#ifndef  attribute
# define attribute(attr)	__attribute__(attr)
#endif
#define alignof(type)		((sizeof(type)+(sizeof(void*)-1)) & ~(sizeof(void*)-1))
#define strsize(string)		((strlen(string)+1)*sizeof(char))
#define MODPREFIX "lookup(udisks): "

typedef struct property_s {
	list_t handle;
	int type;
	char *name;
	void *value;
} property_t;

typedef struct array_s {
	list_t handle;
	char *value;
} array_t;

typedef struct filter_s {
	char *property;
#define DBUS_TYPE_OBJECT DBUS_TYPE_OBJECT_PATH
	int type;
} filter_t;

typedef struct option_s {
	int error;
	union {
		dbus_bool_t	boolean;
		dbus_int32_t	int32;
		dbus_uint32_t	uint32;
		dbus_int64_t	int64;
		dbus_uint64_t	uint64;
		double		number;
		const char	*string;
		const list_t	*array;
	};
} option_t;

typedef struct entry_s {
	time_t   age;
	size_t	 key_len;
	size_t	 mapent_len;
	char	*key;
	char	*mapent;
} entry_t;

typedef struct device_s {
	list_t handle;
	list_t properties;
	list_t *head;
	filter_t *opts;
	entry_t *entry;
	char* identify;
	char* node;
} device_t;

typedef struct session_s {
	list_t handle;
	list_t properties;
	list_t *head;
	filter_t *opts;
	char* identify;
	char* node;
} session_t;

typedef struct mntopt_s {
	list_t handle;
	char* identify;
	char* options;
} mntopt_t;

typedef struct config_s {
	int enabled;
	char* common;
	list_t fstypes;
	list_t byid;
	list_t label;
} config_t;

struct lookup_context {
	pthread_mutex_t mtx;
	pthread_t watchdog;
	dbus_bool_t monitor;
	volatile int active;
	volatile int running;
	const char *mapname;
	list_t devices;
	list_t sessions;
	config_t config;
	DBusConnection *conn;
	DBusError *error;
	struct autofs_point *ap;
	struct map_source *map;
	struct parse_mod *parse;
};

static inline void lock(struct lookup_context *ctxt)
{
	int status = pthread_mutex_lock(&ctxt->mtx);
	if (status)
		fatal(status);
	return;
}

static inline void unlock(struct lookup_context *ctxt)
{
	int status = pthread_mutex_unlock(&ctxt->mtx);
	if (status)
		fatal(status);
	return;
}

/*
 * Use posix_memalign for  a well aligned container which not only has
 * the required amount of space for the structure of a list member but
 * also for the used strings therein.   With this such a container can
 * be freed at once.
 */
extern void* newaligned(size_t size) attribute((__warn_unused_result__,__always_inline__));
extern inline void* newaligned(size_t size)
{
	void *restrict new;
	if (posix_memalign((void**)&new, sizeof(void*), size) != 0) {
		char buf[MAX_ERR_BUF];
		const char *const estr = strerror_r(errno, buf, sizeof(buf));
		logerr(MODPREFIX "memory allocation: %s", estr);
		return (void*)0;
	}
	return new;
}

/*
 * Sorted with same order as found in the response of the dbus call
 * org.freedesktop.DBus.Properties.GetAll, e.g.
 *  dbus-send --system --print-reply --dest=org.freedesktop.UDisks \
 *      /org/freedesktop/UDisks/devices/sda \
 *      org.freedesktop.DBus.Properties.GetAll \
 *      string:org.freedesktop.UDisks.Device | \
 *  sed -rn '/dict entry\(/,/\)/ {
 *      H
 *      x
 *      s@\n*@@g
 *      s@^\s+string\s"(\w+)"\s+.*@\t\1,@p
 *  }'
 *
 * compare with filter[] found in lookup_udisks.c
 */

typedef enum {
	NativePath = 0,
	DeviceDetectionTime,
	DeviceMediaDetectionTime,
	DeviceMajor,
	DeviceMinor,
	DeviceFile,
	DeviceFilePresentation,
	DeviceFileById,
	DeviceFileByPath,
	DeviceIsSystemInternal,
	DeviceIsPartition,
	DeviceIsPartitionTable,
	DeviceIsRemovable,
	DeviceIsMediaAvailable,
	DeviceIsMediaChangeDetected,
	DeviceIsMediaChangeDetectionPolling,
	DeviceIsMediaChangeDetectionInhibitable,
	DeviceIsMediaChangeDetectionInhibited,
	DeviceIsReadOnly,
	DeviceIsDrive,
	DeviceIsOpticalDisc,
	DeviceIsMounted,
	DeviceMountPaths,
	DeviceMountedByUid,
	DeviceIsLuks,
	DeviceIsLuksCleartext,
	DeviceIsLinuxMdComponent,
	DeviceIsLinuxMd,
	DeviceIsLinuxLvm2LV,
	DeviceIsLinuxLvm2PV,
	DeviceIsLinuxDmmpComponent,
	DeviceIsLinuxDmmp,
	DeviceIsLinuxLoop,
	DeviceSize,
	DeviceBlockSize,
	DevicePresentationHide,
	DevicePresentationNopolicy,
	DevicePresentationName,
	DevicePresentationIconName,
	DeviceAutomountHint,
	JobInProgress,
	JobId,
	JobInitiatedByUid,
	JobIsCancellable,
	JobPercentage,
	IdUsage,
	IdType,
	IdVersion,
	IdUuid,
	IdLabel,
	LuksHolder,
	LuksCleartextSlave,
	LuksCleartextUnlockedByUid,
	PartitionSlave,
	PartitionScheme,
	PartitionType,
	PartitionLabel,
	PartitionUuid,
	PartitionFlags,
	PartitionNumber,
	PartitionOffset,
	PartitionSize,
	PartitionAlignmentOffset,
	PartitionTableScheme,
	PartitionTableCount,
	DriveVendor,
	DriveModel,
	DriveRevision,
	DriveSerial,
	DriveWwn,
	DriveRotationRate,
	DriveWriteCache,
	DriveConnectionInterface,
	DriveConnectionSpeed,
	DriveMediaCompatibility,
	DriveMedia,
	DriveIsMediaEjectable,
	DriveCanDetach,
	DriveCanSpindown,
	DriveIsRotational,
	DriveAdapter,
	DrivePorts,
	DriveSimilarDevices,
	OpticalDiscIsBlank,
	OpticalDiscIsAppendable,
	OpticalDiscIsClosed,
	OpticalDiscNumTracks,
	OpticalDiscNumAudioTracks,
	OpticalDiscNumSessions,
	DriveAtaSmartIsAvailable,
	DriveAtaSmartTimeCollected,
	DriveAtaSmartStatus,
	DriveAtaSmartBlob,
	LinuxMdComponentLevel,
	LinuxMdComponentPosition,
	LinuxMdComponentNumRaidDevices,
	LinuxMdComponentUuid,
	LinuxMdComponentName,
	LinuxMdComponentHomeHost,
	LinuxMdComponentVersion,
	LinuxMdComponentHolder,
	LinuxMdComponentState,
	LinuxMdState,
	LinuxMdLevel,
	LinuxMdUuid,
	LinuxMdHomeHost,
	LinuxMdName,
	LinuxMdNumRaidDevices,
	LinuxMdVersion,
	LinuxMdSlaves,
	LinuxMdIsDegraded,
	LinuxMdSyncAction,
	LinuxMdSyncPercentage,
	LinuxMdSyncSpeed,
	LinuxLvm2PVUuid,
	LinuxLvm2PVNumMetadataAreas,
	LinuxLvm2PVGroupName,
	LinuxLvm2PVGroupUuid,
	LinuxLvm2PVGroupSize,
	LinuxLvm2PVGroupUnallocatedSize,
	LinuxLvm2PVGroupSequenceNumber,
	LinuxLvm2PVGroupExtentSize,
	LinuxLvm2PVGroupPhysicalVolumes,
	LinuxLvm2PVGroupLogicalVolumes,
	LinuxLvm2LVName,
	LinuxLvm2LVUuid,
	LinuxLvm2LVGroupName,
	LinuxLvm2LVGroupUuid,
	LinuxDmmpComponentHolder,
	LinuxDmmpName,
	LinuxDmmpSlaves,
	LinuxDmmpParameters,
	LinuxLoopFilename,
	NullDict
} devdict_t;

/*
 * Sorted with same order as found in the response of the dbus call
 * org.freedesktop.DBus.Properties.GetAll, e.g.
 *  dbus-send --print-reply --system --dest=org.freedesktop.ConsoleKit \
 *	/org/freedesktop/ConsoleKit/Session1 \
 *	org.freedesktop.DBus.Properties.GetAll \
 *	string:org.freedesktop.ConsoleKit.Session | \
 *  sed -rn '/dict entry\(/,/\)/{
 *	H
 *	x
 *	s@\n*@@g
 *	s@^\s+string\s"([a-z0-9-]+)"\s+.*@\t\1,@p
 *	}'
 *
 * compare with sessions[] found in lookup_udisks.c
 */

typedef enum {
        unix_user = 0,
        user,
        session_type,
        remote_host_name,
        display_device,
        x11_display,
        x11_display_device,
        active,
        is_local,
        idle_hint,
	NullCk
} ckdict_t;

typedef enum {
	AutofsUdisks = 0,
	MountOptions,
	Common,
	DiskById,
	DiskByLabel,
	FSType,
	NullCnf
} cnfxml_t;

#endif
